import numpy as np
import scipy.signal as signal
import multiprocessing

class FroCorr:
    """An implementation the Frobenius-norm-of-correlation-matricies metric.

    This is not a class to be instantiated, but rather a way to organize and separate the 
    parameterization and comparison steps of the metric calculation to optimize a distance 
    matrix computation (e.g., compute the correlation matrix for each datapoint `once`, then
    just compare correlation matricies).

    """
        
    def parameterize(D):
        """Compute the correlation matrix of a single data point.

        Parameters
        ----------
        D : :obj:`DataSet`
            The lemur data set object to parameterize.

        Returns
        -------
        :obj:`list` of :obj:`ndarray`
            The correlation matrix of each object in the dataset.

        """
        with np.errstate(divide = 'ignore', invalid = 'ignore'):
            return list(map(lambda j: np.nan_to_num(np.corrcoef(D.getMatrix(j))), range(D.n)))

    def compare(x, y):
        """Compute the euclidian distance of two correlation matricies.

        Parameters
        ----------
        x : :obj:`ndarray`
            The left correlation matrix argument.
        y : :obj:`ndarray`
            The left correlation matrix argument.

        Returns
        -------
        float
            The distance.

        """
        return np.linalg.norm(x - y)

class DiffAve:
    """A MRI distance. The euclidian distance of the average voxel intensity volumes.

    This is not a class to be instantiated, but rather a way to organize and separate the 
    parameterization and comparison steps of the metric calculation to optimize a distance 
    matrix computation (e.g., compute the correlation matrix for each datapoint `once`, then
    just compare correlation matricies).

    """
        
    def parameterize(D):
        """Compute the correlation matrix of a single data point.

        Parameters
        ----------
        D : :obj:`DataSet`
            The lemur data set object to parameterize.

        Returns
        -------
        :obj:`list` of :obj:`ndarray`
            The average of each object in the dataset.

        """
        with np.errstate(divide = 'ignore', invalid = 'ignore'):
            return list(map(lambda j: np.mean(D.getMatrix(j), axis=3), range(D.n)))

    def compare(x, y):
        """Compute the euclidian distance of two average volumes.

        Parameters
        ----------
        x : :obj:`ndarray`
            The left correlation matrix argument.
        y : :obj:`ndarray`
            The left correlation matrix argument.

        Returns
        -------
        float
            The distance.

        """
        return np.linalg.norm(x - y)

class NanNorm:
    """The dot product between two vectors, except nans are just treated as 0.

    """
        
    def parameterize(D):
        """Identity function.

        Parameters
        ----------
        D : :obj:`DataSet`
            A dataset.

        Returns
        -------
        :obj:`list` of :obj:`ndarray`
            The a list of each vector in the dataset.

        """

        return list(map(lambda j: D.getResource(j), range(D.n)))

    def compare(x, y):
        """Compute the euclidian distance of two correlation matricies.

        Parameters
        ----------
        x : :obj:`ndarray`
            The left vector argument.
        y : :obj:`ndarray`
            The left vector argument.

        Returns
        -------
        float
            The distance.

        """
        return np.nansum((x - y) * (x - y))

class VectorDifferenceNorm:
        
    def parameterize(D):
        """Identity function.

        Parameters
        ----------
        D : :obj:`DataSet`
            A dataset.

        Returns
        -------
        :obj:`list` of :obj:`ndarray`
            The a list of each vector in the dataset.

        """

        return list(map(lambda j: D.getResource(j), range(D.n)))

    def compare(x, y):
        """Compute the euclidian distance of two correlation matricies.

        Parameters
        ----------
        x : :obj:`ndarray`
            The left vector argument.
        y : :obj:`ndarray`
            The left vector argument.

        Returns
        -------
        float
            The distance.

        """
        return np.linalg.norm((x - y))

class Coh:
    """An implementation of the coherence metric.
    
    This is not a class to be instantiated, but just a method to calculate the intra-datapoint
    coherence distance.

    """
    
    def parameterize(D):
        """Compute the coherence matrix of a single data point.

        Parameters
        ----------
        D : :obj:`ndarray`
            A data matrix on which to compute the coherence matrix.

        Returns
        -------
        :obj:`ndarray`
            The coherence matrix.

        """
        NUM_WORKERS = multiprocessing.cpu_count() - 1
        dat = D.getResource(0)
    
        coherence = np.zeros((dat.shape[0], dat.shape[0]))
        coherence_pars = [(i, j, dat) for i in range(dat.shape[0]) for j in range(i, dat.shape[0])]

        pool = multiprocessing.Pool(processes=NUM_WORKERS)
        results = pool.map_async(Coh.get_coh, coherence_pars)
        coherence_vals = results.get()

        for ((i, j, dat), val) in zip(coherence_pars, coherence_vals):
            coherence[i, j] = val
            coherence[j, i] = val

        return coherence

    def get_coh(tup):
        i, j, dat = tup[0], tup[1], tup[2]
        with np.errstate(divide = 'ignore', invalid = 'ignore'):
            return np.mean(np.nan_to_num(signal.coherence(dat[i, :], dat[j, :], fs=500)[1]))

    def compare(x, y):
        """Compute the euclidian distance of two correlation matricies.

        Parameters
        ----------
        x : :obj:`ndarray`
            The left correlation matrix argument.
        y : :obj:`ndarray`
            The left correlation matrix argument.

        Returns
        -------
        float
            The distance.

        """
        return np.linalg.norm(x - y)




from sklearn.manifold import TSNE, MDS
import pandas as pd
import numpy as np

import lemur.datasets as lds

class BaseEmbedder:
    """A generic embedder object to be extended.

    Parameters
    ----------
    num_components : int
	The number of dimensions the embedding should have.

    Attributes
    ----------
    num_components : int
        The number of dimensions the embedding should have.

    """
    def __init__(self, num_components = 2):
        self.num_components = num_components

class MDSEmbedder(BaseEmbedder):
    embedding_name = "MDS"
    def embed(self, DM):
        """Embed a distance matrix using MDS.

        Parameters
        ----------
        M : :obj:`ndarray`
            The distance matrix to be embedded

        Returns
        -------
        :obj:`ndarray`
            A :obj:`ndarray` of the embedding.

        """
        mds = MDS(n_components = self.num_components, dissimilarity="precomputed")
        mds.fit(DM.getMatrix())
        emb = mds.embedding_
        emb = pd.DataFrame(emb)
        emb.index = DM.D.index
        emb.index.name = DM.D.index.name
        name = DM.DS.name + " " + \
               DM.metric_name + " " + \
               self.embedding_name
        EDS = lds.DataSet(emb, name)
        return EDS

class TSNEEmbedder(BaseEmbedder):
    """A TSNE embedder.

    This uses the sklearn.manifold.TSNE function.

    Attributes
    ----------
    embedding_name : str
        The name of this embedding (default is `TSNE`)

    """
    embedding_name = "TSNE"
    def embed(self, M):
        """Embed a distance matrix using TSNE.

        Parameters
        ----------
        M : :obj:`ndarray`
            The distance matrix to be embedded

        Returns
        -------
        :obj:`ndarray`
            A :obj:`ndarray` of the embedding.

        """
        tsne = TSNE(n_components = self.num_components, metric="precomputed")
        tsne.fit(M)
        emb = tsne.embedding_
        return emb 

def SVD(D):
    """A helper function to do SVD.

    Parameters
    ----------
    D : :obj:`ndarray`
        The matrix to compute SVD on.

    Returns
    -------
    U : :obj:`ndarray`
        An :obj:`ndarray` of the left singular matrix.
    s : :obj:`ndarray`
        An :obj:`ndarray` of the singular values (just a vector, not a square matrix).
    Vt : :obj:`ndarray`
        An :obj:`ndarray` of the right singular matrix.
    m : :obj:`ndarray`
        An :obj:`ndarray` of the affine translation required to center the data  (just a vector, not a square matrix).

    """
    d, n = D.shape
    mu = np.mean(D, axis=1).reshape(d, 1)
    D = D - mu
    U, s, Vt = np.linalg.svd(D, full_matrices=False)
    return U, s, Vt, mu

class PCAEmbedder(BaseEmbedder):
    embedding_name = "PCA"
    def embed(self, M):
        U, s, Vt, m = SVD(M) 
        return (U[:, :self.num_components].T.dot(M)).T  



# qa_graphs.py
# Created by Greg Kiar on 2016-05-11.
# Edited by Eric Bridgeford.
# Edited by Coleman Zhang
# Email: gkiar@jhu.edu

from __future__ import print_function
from argparse import ArgumentParser
from collections import OrderedDict
from scipy.stats import gaussian_kde, rankdata

import numpy as np
import nibabel as nb
import networkx as nx
import pickle
import sys
import os

def loadGraphs(filenames, atlas = None, verb=False):
    """
    Given a list of files, returns a dictionary of graphs
    Required parameters:
        filenames:
            - List of filenames for graphs
    Optional parameters:
        atlas:
            - Name of atlas of interest as it appears in the directory titles
        verb:
            - Toggles verbose output statements
    """
    #  Initializes empty dictionary
    if type(filenames) is not list:
        filenames = [filenames]
    gstruct = OrderedDict()
    for idx, files in enumerate(filenames):
        if verb:
            print("Loading: " + files)
        #  Adds graphs to dictionary with key being filename
        fname = os.path.basename(files)
        try:
            gstruct[fname] = nx.read_graphml(files)
        except:
            gstruct[fname] = nx.read_gpickle(files)
    return gstruct

def compute_metrics(fs, verb=False, modality='dwi'):
    """
    Given a set of files and a directory to put things, loads graphs and
    performs set of analyses on them, storing derivatives in a pickle format
    in the desired output location.
    Required parameters:
        fs:
            - List of filenames (string) in each dataset
    Optional parameters:
        verb:
            - Toggles verbose output statements
        modality:
            - Default: 'dwi'; but can also be 'func'
    Output:
        statsDict:
            - A dictionary of summary statistics of fs
    """
    gr = loadGraphs(fs, verb=verb)
    if modality == 'func':
        graphs = binGraphs(gr)
    else:
        graphs = gr
 
    nodes = nx.number_of_nodes(graphs.values()[0])
    #  Number of non-zero edges (i.e. binary edge count)
    print("Computing: NNZ")
    nnz = OrderedDict((subj, len(nx.edges(graphs[subj]))) for subj in graphs)
    statsDict = {'number_non_zeros': nnz}
    print("Sample Mean: %.2f" % np.mean(nnz.values()))

    # Scan Statistic-1
    print("Computing: Max Local Statistic Sequence")
    temp_ss1 = scan_statistic(graphs, 1)
    ss1 = temp_ss1
    statsDict['locality_statistic'] = ss1
    show_means(temp_ss1)

    if modality == 'func':
        graphs = rankGraphs(gr)
        wt_args = {'weight': 'weight'}
    else:
        wt_args = {}

    #   Clustering Coefficients
    print("Computing: Clustering Coefficient Sequence")
    temp_cc = OrderedDict((subj, nx.clustering(graphs[subj],
                                               **wt_args).values())
                          for subj in graphs)
    ccoefs = temp_cc
    statsDict['clustering_coefficients'] = ccoefs
    show_means(temp_cc)

    #  Degree sequence
    print("Computing: Degree Sequence")
    total_deg = OrderedDict((subj, np.array(nx.degree(graphs[subj],
                                                      **wt_args).values()))
                            for subj in graphs)
    ipso_deg = OrderedDict()
    contra_deg = OrderedDict()
    for subj in graphs:  # TODO GK: remove forloop and use comprehension maybe?
        g = graphs[subj]
        N = len(g.nodes())
        LLnodes = g.nodes()[0:N/2]  # TODO GK: don't assume hemispheres
        LL = g.subgraph(LLnodes)
        LLdegs = [LL.degree(**wt_args)[n] for n in LLnodes]

        RRnodes = g.nodes()[N/2:N]  # TODO GK: don't assume hemispheres
        RR = g.subgraph(RRnodes)
        RRdegs = [RR.degree(**wt_args)[n] for n in RRnodes]

        LRnodes = g.nodes()
        ipso_list = LLdegs + RRdegs
        degs = [g.degree(**wt_args)[n] for n in LRnodes]
        contra_deg[subj] = [a_i - b_i for a_i, b_i in zip(degs, ipso_list)]
        ipso_deg[subj] = ipso_list
        # import pdb; pdb.set_trace()

    deg = {'total_deg': total_deg,
           'ipso_deg': ipso_deg,
           'contra_deg': contra_deg}
    statsDict['degree_distribution'] = deg
    show_means(total_deg)

    #  Edge Weights
    if modality == 'dwi':
        print("Computing: Edge Weight Sequence")
        temp_ew = OrderedDict((s, [graphs[s].get_edge_data(e[0], e[1])['weight']
                               for e in graphs[s].edges()]) for s in graphs)
        ew = temp_ew
        statsDict['edge_weight'] = ew
        show_means(temp_ew)
    else:
        temp_pl = OrderedDict()
        print("Computing: Path Length Sequence")
        nxappl = nx.all_pairs_dijkstra_path_length
        for s in graphs:
            apd = nxappl(graphs[s])
            # iterate over the nodes to find the average distance to each node
            avg_path = [np.nanmean(v.values()) for k, v in apd.iteritems()]
            temp_pl[s] = np.array(avg_path)
        pl = temp_pl
        statsDict['path_length'] = pl
        show_means(pl)

    # Eigen Values
    print("Computing: Eigen Value Sequence")
    laplac = OrderedDict((subj, nx.normalized_laplacian_matrix(graphs[subj]))
                         for subj in graphs)
    eigs = OrderedDict((subj, np.sort(np.linalg.eigvals(laplac[subj].A))[::-1])
                       for subj in graphs)
    statsDict['eigen_sequence'] = eigs
    print("Subject Maxes: " + ", ".join(["%.2f" % np.max(eigs[key])
                                         for key in eigs.keys()]))

    # Betweenness Centrality
    print("Computing: Betweenness Centrality Sequence")
    nxbc = nx.algorithms.betweenness_centrality  # For PEP8 line length...
    temp_bc = OrderedDict((subj, nxbc(graphs[subj], **wt_args).values())
                          for subj in graphs)
    centrality = temp_bc
    statsDict['betweenness_centrality'] = centrality
    show_means(temp_bc)

    # Mean connectome
    print("Computing: Mean Connectome")
    nxnp = nx.to_numpy_matrix
    adj = OrderedDict((subj, nxnp(graph, nodelist=sorted(graph.nodes())))
                      for subj, graph in graphs.iteritems())
    mat = np.zeros(adj.values()[0].shape)
    for subj in adj:
        mat += adj[subj]
    mat = mat/len(adj.keys())
    statsDict['study_mean_connectome'] = mat
    return statsDict


def show_means(data):
    print("Subject Means: " + ", ".join(["%.2f" % np.mean(data[key])
                                         for key in data.keys()]))

def binGraphs(graphs, thr=0.1):
    """
    Binarizes a set of graphs given a threshold.
    Required Parameters:
        graphs:
            - a list of graphs.
        thr:
            - the threshold below which edges will be assumed disconnected.
              .1 is chosen as default according to discriminability results.
    """
    binGraphs = {}
    for subj, graph in graphs.iteritems():
        bin_graph = nx.Graph()
        for (u, v, d) in graph.edges(data=True):
            if d['weight'] > thr:
                bin_graph.add_edge(u, v, weight=1)
        binGraphs[subj] = bin_graph
    return binGraphs


def rankGraphs(graphs):
    """
    Ranks the edges in each element of a list of graphs.
    Required Parameters:
        graphs:
            - a list of graphs.
    """
    rankGraphs = {}
    for subj, graph in graphs.iteritems():
        rgraph = nx.Graph()
        edge_ar = np.asarray([x[2]['weight'] for x in graph.edges(data=True)])
        rank_edge = rankdata(edge_ar)  # rank the edges
        for ((u, v, d), rank) in zip(graph.edges(data=True), rank_edge):
            rgraph.add_edge(u, v, weight=rank)
        rankGraphs[subj] = rgraph
    return rankGraphs


def scan_statistic(mygs, i):
    """
    Computes scan statistic-i on a set of graphs
    Required Parameters:
        mygs:
            - Dictionary of graphs
        i:
            - which scan statistic to compute
    """
    ss = OrderedDict()
    for key in mygs.keys():
        g = mygs[key]
        tmp = np.array(())
        for n in g.nodes():
            sg = nx.ego_graph(g, n, radius=i)
            tmp = np.append(tmp, np.sum([sg.get_edge_data(e[0], e[1])['weight']
                            for e in sg.edges()]))
        ss[key] = tmp
    return ss


def density(data, nbins=500, rng=None):
    """
    Computes density for metrics which return vectors
    Required parameters:
        data:
            - Dictionary of the vectors of data
    """
    density = OrderedDict()
    xs = OrderedDict()
    for subj in data:
        hist = np.histogram(data[subj], nbins)
        hist = np.max(hist[0])
        dens = gaussian_kde(data[subj])
        if rng is not None:
            xs[subj] = np.linspace(rng[0], rng[1], nbins)
        else:
            xs[subj] = np.linspace(0, np.max(data[subj]), nbins)
        density[subj] = dens.evaluate(xs[subj])*np.max(data[subj]*hist)
    return {"xs": xs, "pdfs": density}


def main():
    """
    Argument parser and directory crawler. Takes organization and atlas
    information and produces a dictionary of file lists based on datasets
    of interest and then passes it off for processing.
    Required parameters:
        atlas:
            - Name of atlas of interest as it appears in the directory titles
        basepath:
            - Basepath for which data can be found directly inwards from
        outdir:
            - Path to derivative save location
    Optional parameters:
        fmt:
            - Determines file organization; whether graphs are stored as
              .../atlas/dataset/graphs or .../dataset/atlas/graphs. If the
              latter, use the flag.
        verb:
            - Toggles verbose output statements
    """
    parser = ArgumentParser(description="Computes Graph Metrics")
    parser.add_argument("atlas", action="store", help="atlas directory to use")
    parser.add_argument("indir", action="store", help="base directory loc")
    parser.add_argument("-f", "--fmt", action="store_true", help="Formatting \
                        flag. True if bc1, False if greg's laptop.")
    parser.add_argument("-v", "--verb", action="store_true", help="")
    result = parser.parse_args()

    #  Sets up directory to crawl based on the system organization you're
    #  working on. Which organizations are pretty clear by the code, methinks..
    indir = result.indir
    atlas = result.atlas

    #  Crawls directories and creates a dictionary entry of file names for each
    #  dataset which we plan to process.
    fs = [indir + "/" + fl
          for root, dirs, files in os.walk(indir)
          for fl in files
          if fl.endswith(".graphml") or fl.endswith(".gpickle")]
    #  The fun begins and now we load our graphs and process them.
    compute_metrics(fs, atlas, result.verb)


if __name__ == "__main__":
    main()


import numpy as np

def gen_samples(params, dists):
    points = dict()
    for dist in dists:
        points.update(_make_dist(params, dist))
    return points

def _make_dist(p_global, p_local):
    N = p_local.get('num_samples', p_global['num_samples'])
    if p_local['type'] == 'gaussian':
        x, y = np.random.multivariate_normal(p_local['mean'], p_local['cov'], N).T
    if 'transforms' in p_local.keys():
        if type(p_local['transforms']) != 'list':
            p_local['transforms'] = [p_local['transforms']]
        for transform in p_local['transforms']:
            if transform == 'rotate':
                x, y = _rotate_points(x, y, p_local['theta'])
    return {p_local['name']: (x, y)}



from abc import ABCMeta, abstractmethod
import numpy as np
from sklearn.mixture import GaussianMixture
import sklearn.cluster as skcl


class Clustering(metaclass=ABCMeta):
    def __init__(self, DS, levels=1, random_state=None):
        self.DS = DS
        self.name = DS.name
        self.columns = DS.D.columns
        self.X = self.DS.D.as_matrix()
        self.levels = levels
        self.random_state = random_state
        self.clusters = []

    @abstractmethod
    def cluster(self):
        pass

class KMeans(Clustering):
    def __init__(self, DS, n_clusters=8, random_state=None):
        """
        Parameters
        ----------
        DS :obj:`Dataset`
        level : int
            Number of levels to cluster
        random_state : int (optional)
            Initialize Gaussian Mixture Model with specified random state
        """
        Clustering.__init__(self, DS, 1, random_state)
        self.n_clusters = n_clusters
        self.clustname = 'KMeans'
        self.shortclustname = 'km'

    def cluster(self):
        self.clusters = [[self.X]]
        clust = skcl.KMeans(n_clusters=self.n_clusters,random_state=self.random_state)
        y = clust.fit_predict(self.X)
        self.clusters.append([(self.X[y == i, :]) for i in range(self.n_clusters)])


class HGMMClustering(Clustering):
    def __init__(self, DS, levels=1, random_state=None):
        """
        Parameters
        ----------
        DS :obj:`Dataset`
        level : int
            Number of levels to cluster
        random_state : int (optional)
            Initialize Gaussian Mixture Model with specified random state
        """
        Clustering.__init__(self, DS, levels, random_state)
        self.clustname = 'HGMM'
        self.shortclustname = 'hgmm'

    def cluster(self):
        clusters = []
        n = self.X.shape[0]
        l0 = self.hgmml0(self.X, self.random_state)
        clusters.append(l0)
        li = self.gmmBranch(l0[0], self.random_state)
        clusters.append(li)
        while (len(li) < n) and (len(clusters) - 1 < self.levels):
            lip = []
            for c in li:
                q = self.gmmBranch(c, self.random_state)
                if q is not None:
                    lip.extend(q)
            clusters.append(lip)
            li = lip

        self.clusters = [list(map(lambda x: x[0], c)) for c in clusters]
        self.hierarch = clusters

    def gmmBranch(self, level, random_state):
        X, p, mu = level

        #Check BIC to see to split node
        gmm_1 = GaussianMixture(n_components=1, random_state=random_state)
        gmm_1.fit(X)
        bic_1 = gmm_1.bic(X)

        if len(X) != 0: #Does not run when input has one sample point
            gmm_2 = GaussianMixture(n_components=2, random_state=random_state)
            gmm_2.fit(X)
            bic_2 = gmm_2.bic(X)
        else:
            bic_2 = bic_1

        if bic_2 < bic_1:
            X0 = X[gmm_2.predict(X) == 0, :]
            X1 = X[gmm_2.predict(X) == 1, :]
            mypro = np.rint(gmm_2.weights_ * p)
            return [(
                X0,
                int(mypro[0]),
                gmm_2.means_[0, :],
            ), (
                X1,
                int(mypro[1]),
                gmm_2.means_[1, :],
            )]
        else:
            return [(
                X,
                int(np.rint(p * gmm_1.weights_[0])),
                gmm_1.means_[0, :],
            )]

    def hgmml0(self, X, random_state):
        gmm = GaussianMixture(n_components=1, random_state=random_state)
        gmm.fit(X)
        return [(
            X,
            int(np.rint(X.shape[0] * gmm.weights_[0])),
            gmm.means_[0, :],
        )]


# qa_graphs_plotting.py
# Created by Greg Kiar on 2016-09-19.
# Email: gkiar@jhu.edu
# Edited by Coleman Zhang

from argparse import ArgumentParser
from plotly.offline import download_plotlyjs, init_notebook_mode, iplot, plot
from . import plotly_helper as pp
import pickle
import numpy as np
import os

def make_panel_plot(statsDict, outf, dataset=None, atlas = None, minimal=True, 
                    log=True, hemispheres=True, modality='dwi'):

    """
    Given a dictionary of statistics generated by qa_graphs.compute_metrics(), 
    output an html file of plotly visualization.
    Required parameters:
        statsDict:
            - dictionary of statistics generated by qa_graphs.compute_metrics()
        outf:
            - The name of the generated plotly html file
    Optional parameters:
        atlas:
            - Name of atlas of interest as it appears in the directory titles
    """    

    dictKeys = list(statsDict.keys())
    dictKeys = sorted(dictKeys, key = str)
    
    if modality == 'dwi':
        modtit = "DWI"
        order = [0, 1, 2, 5, 4, 3, 6, 7]
        dictKeys = [dictKeys[o] for o in order]
        labs = ['Betweenness Centrality', 'Clustering Coefficient', 'Degree',
                'Locality Statistic-1', 'Eigenvalue', 'Edge Weight',
                'Number of Non-zeros', 'Mean Connectome']
        
    else:
        modtit = "fMRI"
        order = [0, 1, 2, 6, 4, 3, 5, 7]
        dictKeys = [dictKeys[o] for o in order]
        labs = ['Betweenness Centrality', 'Clustering Coefficient', 'Degree',
                'Average Path Length', 'Locality Statistic-1', 'Eigenvalue',
                'Number of Non-zeros', 'Mean Connectome']
    traces = list(())
    
    for idx, curr in enumerate(dictKeys):
        dat = statsDict[dictKeys[idx]]
        if dictKeys[idx] == 'number_non_zeros':
            fig = pp.plot_rugdensity(dat.values())
        elif dictKeys[idx] == 'edge_weight':
            edges = np.max([len(dat[i]) for i in dat.keys()])
            fig = pp.plot_series(dat.values(), sort=True)
        elif dictKeys[idx] == 'degree_distribution':
            fig = pp.plot_degrees(dat, hemi=hemispheres)
            if hemispheres:
                maxdat = np.max([np.max(dat[key][k]) 
                                 for key in dat.keys()
                                 for k in dat[key]])
                anno = [dict(x=dims/3,
                             y=4*float(maxdat/7),
                             xref='x3',
                             yref='y3',
                             text='ipsilateral',
                             showarrow=False,
                             font=dict(color='rgba(0.0,0.0,0.0,0.6)',
                                       size=14)),
                        dict(x=dims/3,
                             y=3.7*float(maxdat/7),
                             xref='x3',
                             yref='y3',
                             text='contralateral',
                             showarrow=False,
                             font=dict(color='rgba(0.11,0.62,0.47,0.6)',
                                       size=14))]
        elif dictKeys[idx] == 'study_mean_connectome':
            if log:
                dat = np.log10(dat+1)
            fig = pp.plot_heatmap(dat, name=labs[idx])
        else:
            dims = len(dat.values()[0])
            fig = pp.plot_series(dat.values())
        traces += [fig_to_trace(fig)]

    multi = pp.traces_to_panels(traces)
    for idx, curr, in enumerate(dictKeys):
        key = 'axis%d' % (idx+1)
        d = multi.layout['x'+key]['domain']
        multi.layout['x'+key]['domain'] = [d[0], d[1]-0.0125]
        multi.layout['x'+key]['zeroline'] = False
        multi.layout['y'+key]['zeroline'] = False
        multi.layout['y'+key]['title'] = ''
        multi.layout['x'+key]['title'] = 'Node'
        multi.layout['x'+key]['nticks'] = 3
        multi.layout['y'+key]['nticks'] = 3
        if (idx in [0, 1, 2, 3, 4, 5] and modality == 'func') or (idx in [0, 1, 2, 3, 4, 5] and modality == 'dwi'):
            multi.layout['x'+key]['range'] = [1, dims]
            multi.layout['x'+key]['tickvals'] = [1, dims/2, dims]
            if idx in [2]:
                if hemispheres:
                    multi.layout['annotations'] = anno
            elif log:
                multi.layout['y'+key]['type'] = 'log'
                multi.layout['y'+key]['title'] += 'log'
        if idx in [5] and modality == 'dwi':
            multi.layout['x'+key]['range'] = [1, edges]
            multi.layout['x'+key]['tickvals'] = [1, edges/2, edges]
            multi.layout['x'+key]['title'] = 'Edge'
        if (idx in [4] and modality == 'dwi') or (idx in [5] and modality == 'func'):
            multi.layout['x'+key]['range'] = [1, dims]
            multi.layout['x'+key]['tickvals'] = [1, dims/2, dims]
            multi.layout['x'+key]['title'] = 'Dimension'
        multi.layout['y'+key]['title'] += labs[idx]
        if (idx in [6]):
            multi.layout['y'+key]['title'] = 'Relative Probability'
            multi.layout['x'+key]['title'] = labs[idx]
        if idx in [7]:
            multi.layout['y'+key]['title'] = None
            multi.layout['x'+key]['title'] = labs[idx]
            multi.layout['y'+key]['autorange'] = 'reversed'
            multi.layout['x'+key]['tickvals'] = [0, dims/2-1, dims-1]
            multi.layout['y'+key]['tickvals'] = [0, dims/2-1, dims-1]
            multi.layout['x'+key]['ticktext'] = [1, dims/2, dims]
            multi.layout['y'+key]['ticktext'] = [1, dims/2, dims]
            if log:
                multi.layout['x'+key]['title'] += ' (log10)'
    if dataset is not None and atlas is not None:
        if atlas == 'desikan':
            atlas = atlas.capitalize()
        tit = "{} Dataset ({} parcellation), {} Group Analysis".format(dataset,
                atlas, modtit)
    else:
        tit = "{} Group Analysis".format(modtit)
    multi.layout['title'] = tit
    # iplot(multi, validate=False)

    if minimal:
        locs = [idx for idx, d in enumerate(multi.data) if d['yaxis'] == 'y8']
        for l in locs:
            multi.data[l] = {}
        multi.layout['x'+key]['title'] = ''
        multi.layout['y'+key]['title'] = ''
        multi = pp.panel_invisible(multi, 8)

    plot(multi, validate=False, filename=outf+'.html')


import os
import imageio
from PIL import ImageDraw, Image, ImageFont

from plotly.offline import iplot, plot
import plotly.graph_objs as go
import plotly.figure_factory as ff
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import hashlib
from ipywidgets import interact
import random
import scipy.signal as signal
import colorlover as cl

from nilearn import image as nimage
from nilearn import plotting as nilplot
import nibabel as nib

from lemur import embedders as leb
from .utils import qa_graphs as qg
from .utils import qa_graphs_plotting as qgp

def get_spaced_colors(n):
    max_value = 255
    interval = int(max_value / n)
    hues = range(0, max_value, interval)
    return cl.to_rgb(["hsl(%d,100%%,40%%)"%i for i in hues])

def get_heat_colors(n):
    max_value = 255
    interval = int(max_value / n)
    hues = range(0, max_value, interval)
    return cl.to_rgb(["hsl(%d,100%%,40%%)"%i for i in hues])

class MatrixPlotter:
    def __init__(self, DS, mode="notebook", base_path = None):
        self.DS = DS
        self.plot_mode = mode
        self.base_path = base_path

    def makeplot(self, fig, local_path=None):
        """Make the plotly figure visable to the user in the way they want.

        Parameters
        ----------
        gid : :obj:`figure`
            An plotly figure.

        """

        if self.plot_mode == "notebook":
            iplot(fig)
        if self.plot_mode == "savediv":
            fig["layout"]["autosize"] = True
            div = plot(fig, output_type='div', include_plotlyjs=False)
            path = os.path.join(self.base_path, local_path + ".html")
            os.makedirs("/".join(path.split("/")[:-1]), exist_ok=True)
            with open(path, "w") as f:
                f.write(div)
                f.close()

        if self.plot_mode == "div":
            fig["layout"]["autosize"] = True
            return plot(fig, output_type='div', include_plotlyjs=False)

class TimeSeriesPlotter:
    """A generic one-to-one plotter for time series data to be extended.

    Parameters
    ----------
    data : :obj:`ndarray`
        The time series data.
    resource_name : string
        The name of the time series being plotted.
    row_name : string
        The name of the rows in the time-series (e.g. channels, sources. ect.).
    column_name : string
        The name of the columns in the time-series (e.g. time points, time steps, seconds, ect.).

    Attributes
    ----------
    data : :obj:`ndarray`
        The time series data.
    d : int
        The number of dimensions in the time series
    n : int
        The number of time points in the time series
    row_name : string
        The name of the rows in the time-series (e.g. channels, sources. ect.).
    column_name : string
        The name of the columns in the time-series (e.g. time points, time steps, seconds, ect.).
    resource_name : string
        The name of the time series being plotted.

    """

    def __init__(self, DS, mode = "notebook", base_path=None):
        self.data = DS.D.as_matrix().T
        self.d, self.n = self.data.shape
        self.row_name = "Channels"
        self.col_name = "Time Points"
        self.resource_name = DS.name
        self.plot_mode = mode
        self.base_path = base_path

    def makeplot(self, fig, local_path=None):
        """Make the plotly figure visable to the user in the way they want.

        Parameters
        ----------
        gid : :obj:`figure`
            An plotly figure.

        """

        if self.plot_mode == "notebook":
            iplot(fig)
        if self.plot_mode == "savediv":
            fig["layout"]["autosize"] = True
            div = plot(fig, output_type='div', include_plotlyjs=False)
            path = os.path.join(self.base_path, local_path + ".html")
            os.makedirs("/".join(path.split("/")[:-1]), exist_ok=True)
            with open(path, "w") as f:
                f.write(div)
                f.close()

        if self.plot_mode == "div":
            fig["layout"]["autosize"] = True
            return plot(fig, output_type='div', include_plotlyjs=False)


class GraphPlotter:
    """
    Visualization of summary statistics of a list of graphs

    Parameters
    ----------
    fs : list
        List of strings of filenames of graphs to be analyzed
    outf: string
        name of plotly html output file
    """
    def __init__(self, fs, outf):

        self.file_list = fs
        self.outf = outf

    def makeplot(self, modality='dwi', atlas = None, log = True):

        statsDict = qg.compute_metrics(fs, modality)
        qpg.make_panel_plot(statsDict, outf, atlas, log, modality)


class SquareHeatmap(MatrixPlotter):
    titlestring = "%s Heatmap"
    shortname = "squareheat"

    def plot(self):
        """Constructs a distance matrix heatmap using the :obj:`DistanceMatrix` object, in plotly.

        """
        title = self.titlestring % (self.DS.name)
        xaxis = go.XAxis(
                title=self.DS.D.index.name,
                ticktext = self.DS.D.index,
                ticks = "",
                showticklabels=False,
                showgrid=False,
                mirror=True,
                tickvals = [i for i in range(len(self.DS.D.index))])
        yaxis = go.YAxis(
                scaleanchor="x",
                title=self.DS.D.index.name,
                ticktext = self.DS.D.index,
                showgrid=False,
                ticks = "",
                showticklabels=False,
                mirror=True,
                tickvals = [i for i in range(len(self.DS.D.index))])
        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)
        trace = go.Heatmap(z = self.DS.D.as_matrix().T)
        data = [trace]
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)

class Scatterplot(MatrixPlotter):
    titlestring = "%s Scatterplot"
    shortname = "scatter"

    def plot(self):
        """Constructs a distance matrix heatmap using the :obj:`DistanceMatrix` object, in plotly.

        """
        title = self.titlestring % (self.DS.name)
        if self.DS.d != 2:
            print("Scatter plot must get 2 dimensional dataset")
            return
        xaxis = go.XAxis(title=self.DS.D.index.name)
        yaxis = go.YAxis(scaleanchor="x", title=self.DS.D.index.name)
        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)
        trace = go.Scatter(x = self.DS.D.as_matrix()[:, 0],
                           y = self.DS.D.as_matrix()[:, 1],
                           mode = "markers")
        data = [trace]
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)

class ThreeDScatterplot(MatrixPlotter):
    titlestring = "%s 3D Scatter Plot"
    shortname = "3sp"

    def plot(self):
        title = self.titlestring % (self.DS.name)

        if self.DS.d < 3:
            print("Scatter plot must get at least 3 dimensional dataset")
            return
        xaxis = go.XAxis(title=list(self.DS.D)[0])
        yaxis = go.YAxis(title=list(self.DS.D)[1])
        zaxis = go.ZAxis(title=list(self.DS.D)[2])
        layout = dict(title=title, scene=dict(
            xaxis=xaxis,
            yaxis=yaxis,
            zaxis=zaxis
        ))
        trace = go.Scatter3d(
            x=self.DS.D.as_matrix()[:, 0],
            y=self.DS.D.as_matrix()[:, 1],
            z=self.DS.D.as_matrix()[:, 2],
            mode='markers',
            marker=dict(
                size=4
            )
        )
        data = [trace]
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)

class SpatialConnectivity(MatrixPlotter):
    titlestring = "%s Spatial Connectivity"
    shortname = "spatialconn"

    def plot(self, spatial):
        title = self.titlestring % (self.DS.name)
        DM = self.DS.D.as_matrix()
        sp = spatial.D.as_matrix()
        trace1 = go.Scatter3d(
            x=sp[:, 0],
            y=sp[:, 1],
            z=sp[:, 2],
            mode='markers',
            marker=dict(
                size=12,
                opacity=0
            )
        )
        #med = np.nanmedian(DM)
        #mask = np.nan_to_num(DM) > (3 * med)
        #Xe = []
        #Ye = []
        #Ze = []
        #for i in range(DM.shape[0]):
        #    for j in range(DM.shape[1]):
        #        if mask[i, j]:
        #            Xe += [sp[i, 0], sp[j, 0], None]
        #            Ye += [sp[i, 1], sp[j, 1], None]
        #            Ze += [sp[i, 2], sp[j, 2], None]
        #print(Xe)
        #trace2 = go.Scatter3d(x=Xe,
        #                      y=Ye,
        #                      z=Ze,
        #                      mode='lines',
        #                      line=go.Line(color='rgb(125,125,125)', width=1),
        #                      hoverinfo='none')
        data = [trace1]
        fig = dict(data=data, layout={"autosize":True})
        return self.makeplot(fig, self.DS.name + "/" + self.shortname)

class ConnectedScatterplot(MatrixPlotter):
    titlestring = "%s Scatterplot"
    shortname = "connectedscatter"

    def plot(self, spatialDM):
        """Constructs a distance matrix heatmap using the :obj:`DistanceMatrix` object, in plotly.

        """
        title = self.titlestring % (self.DS.name)
        DM = self.DS.D.as_matrix()
        sDM = spatialDM.D.as_matrix()
        colors = (np.nansum(DM, axis=0) - 1) / DM.shape[0]
        TSNEEmbedder = leb.TSNEEmbedder(num_components=2)
        m = TSNEEmbedder.embed(sDM)
        xaxis = go.XAxis(title=spatialDM.D.index.name)
        yaxis = go.YAxis(scaleanchor="x", title=spatialDM.D.index.name)
        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis, showlegend=False)
        trace1 = go.Scatter(x = m[:, 0],
                           y = m[:, 1],
                           mode = "markers",
                           marker = dict(color=colors, size=16, showscale=True))
        distances = np.zeros([m.shape[0], m.shape[0]])
        for i in range(m.shape[0]):
            for j in range(m.shape[0]):
                distances[i, j] = np.linalg.norm(m[i, :] - m[j, :])
        med = np.median(distances)
        mask = distances < (med / 4)
        Xe = []
        Ye = []
        for i in range(distances.shape[0]):
            for j in range(distances.shape[1]):
                if mask[i, j]:
                    Xe += [m[i, 0], m[j, 0], None]
                    Ye += [m[i, 1], m[j, 1], None]
        trace2 = go.Scatter(x=Xe,
                            y=Ye,
                            mode='lines',
                            line=go.Line(color='rgb(125,125,125)', width=1),
                            hoverinfo='none')
        data = [trace1, trace2]
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, self.DS.name + "/" + self.shortname)

class Scatterplot(MatrixPlotter):
    titlestring = "%s Scatterplot"
    shortname = "scatter"

    def plot(self):
        """Constructs a distance matrix heatmap using the :obj:`DistanceMatrix` object, in plotly.

        """
        title = self.titlestring % (self.DS.name)
        if self.DS.d != 2:
            print("Scatter plot must get 2 dimensional dataset")
            return
        xaxis = go.XAxis(title=self.DS.D.index.name)
        yaxis = go.YAxis(scaleanchor="x", title=self.DS.D.index.name)
        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)
        trace = go.Scatter(x = self.DS.D.as_matrix()[:, 0],
                           y = self.DS.D.as_matrix()[:, 1],
                           mode = "markers")
        data = [trace]
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)

class Heatmap(MatrixPlotter):
    titlestring = "%s Heatmap"
    shortname = "heatmap"

    def plot(self):
        title = self.titlestring % (self.DS.name)
        xaxis = go.XAxis(
                title="observaions",
                ticktext = self.DS.D.index,
                showticklabels=False,
                tickvals = [i for i in range(len(self.DS.D.index))])
        yaxis = go.YAxis(
                title="dimensions",
                ticktext = self.DS.D.columns,
                showticklabels=False,
                tickvals = [i for i in range(len(self.DS.D.columns))])
        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)
        trace = go.Heatmap(z = self.DS.D.as_matrix().T)
        data = [trace]
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)

class LocationHeatmap(MatrixPlotter):
    titlestring = "%s Location Heatmap"
    shortname = "locationheat"

    def plot(self, showticklabels=False):
        title = self.titlestring % (self.DS.name)
        D = self.DS.D.as_matrix().T
        means = np.mean(D, axis=1)
        medians = np.median(D, axis=1)
        z = np.vstack([means, medians])
        yaxis = go.YAxis(
                ticktext = ["mean", "median"],
                showticklabels=True,
                tickvals = [0, 1])
        xaxis = go.XAxis(
                title="dimensions",
                showticklabels=showticklabels)
        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)
        trace = go.Heatmap(x = self.DS.D.columns,
                           z = z)
        data = [trace]
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)

class LocationLines(MatrixPlotter):
    titlestring = "%s Embedding Location Lines"
    shortname = "locationlines"

    def plot(self, showticklabels=False):
        title = self.titlestring % (self.DS.name)
        D = self.DS.D.as_matrix().T
        means = np.mean(D, axis=1)
        medians = np.median(D, axis=1)
        trace0 = go.Scatter(x = self.DS.D.columns,
                            y = means,
                            name="means")
        trace1 = go.Scatter(x = self.DS.D.columns,
                            y = medians,
                            name="medians")
        layout = dict(title=title,
                      xaxis=dict(title="Dimensions",
                                 showticklabels=showticklabels),
                      yaxis=dict(title="Mean or Median Value"))
        data = [trace0, trace1]
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)

class HistogramHeatmap(MatrixPlotter):
    titlestring = "%s Histogram Heatmap"
    shortname = "histogramheat"

    def plot(self, showticklabels=False):
        title = self.titlestring % (self.DS.name)
        D = self.DS.D.as_matrix().T
        d, n = D.shape
        D = (D - np.mean(D, axis=1).reshape(d, 1)) / np.std(D, axis=1).reshape(d, 1)
        D = np.nan_to_num(D) # only nan if std all 0 -> all values 0
        num_bins = int(np.sqrt(n))
        bins = np.linspace(-5, 5, num_bins + 1)
        bin_centers = (bins[1:] + bins[:-1]) / 2
        H = []
        for i in range(D.shape[0]):
            hist = np.histogram(D[i, :], bins = bins)[0]
            H.append(hist)
        z = np.vstack(H)
        trace = go.Heatmap(y = self.DS.D.columns,
                           z = z)
        data = [trace]
        xaxis = go.XAxis(
                title="Normalized Value",
                ticktext = list(map(lambda x: "%0.4f"%x, bin_centers)),
                ticks = "",
                showticklabels=True,
                mirror=True,
                tickvals = [i for i in range(len(bin_centers))])
        yaxis = go.YAxis(
                title="Dimensions",
                ticks = "",
                showticklabels=showticklabels,
                mirror=True)
        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)
        fig = dict(data = data, layout = layout)
        return self.makeplot(fig, "agg/" + self.shortname)

class CorrelationMatrix(MatrixPlotter):
    titlestring = "%s Correlation Matrix"
    shortname = "correlation"

    def plot(self, showticklabels=False):
        title = self.titlestring % (self.DS.name)
        D = self.DS.D.as_matrix().T
        xaxis = dict(
            title = "Dimensions",
            ticks = "",
            showgrid=False,
            zeroline=False,
            showticklabels=showticklabels,
        )
        yaxis = dict(
            scaleanchor="x",
            title = "Dimensions",
            ticks = "",
            showgrid=False,
            zeroline=False,
            showticklabels=showticklabels,
        )
        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)
        with np.errstate(divide = 'ignore', invalid = 'ignore'):
            C = np.nan_to_num(np.corrcoef(D))

        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)
        trace = go.Heatmap(x = self.DS.D.columns,
                           y = self.DS.D.columns,
                           z = C,
                           zmin = -1,
                           zmax = 1)
        fig = dict(data=[trace], layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)

class ScreePlotter(MatrixPlotter):
    titlestring = "%s Scree Plot"
    shortname = "scree"

    def plot(self):
        title = self.titlestring % (self.DS.name)
        D = self.DS.D.as_matrix().T
        _, S, _ = np.linalg.svd(D, full_matrices=False)
        y = S
        x = np.arange(1, len(S) + 1)
        sy = np.sum(y)
        cy = np.cumsum(y)
        xaxis = dict(
            title = 'Factors'
        )
        yaxis = dict(
            title = 'Proportion of Total Variance'
        )
        var = go.Scatter(mode = 'lines+markers',
                         x = x,
                         y = y / sy,
                         name = "Variance")
        cumvar = go.Scatter(mode = 'lines+markers',
                            x = x,
                            y = cy / sy,
                            name = "Cumulative Variance")
        data = [var, cumvar]
        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)

class EigenvectorHeatmap(MatrixPlotter):
    titlestring = "%s Eigenvector Heatmap"
    shortname = "evheat"

    def plot(self):
        title = self.titlestring % (self.DS.name)
        D = self.DS.D.as_matrix().T
        d, n = D.shape
        U, _, _ = np.linalg.svd(D, full_matrices=False)
        xaxis = go.XAxis(
                title="Eigenvectors",
                ticktext = ["Eigenvector %s"%i for i in range(1, d + 1)],
                ticks = "",
                showgrid=False,
                showticklabels=False,
                mirror=True,
                tickvals = [i for i in range(d)])
        yaxis = go.YAxis(
                title="Eigenvector Components",
                scaleanchor="x",
                showgrid=False,
                ticktext = ["Component %s"%i for i in range(1, d + 1)],
                ticks = "",
                showticklabels=False,
                mirror=True,
                tickvals = [i for i in range(d)])
        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)
        trace = go.Heatmap(z = U)
        data = [trace]
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)

class HierarchicalClusterMeansDendrogram(MatrixPlotter):

    titlestring = "%s %s Cluster Means Dendrogram, Level %d"
    shortname = "cmd"

    def plot(self):
        title = self.titlestring % (self.DS.name, self.DS.clustname, self.DS.levels)
        self.shortname = self.DS.shortclustname + self.shortname
        means = []
        for c in self.DS.clusters[self.DS.levels]:
            means.append(np.average(c, axis=0))
        X = np.column_stack(means).T
        fig = ff.create_dendrogram(X)
        fig["layout"]["title"] = title
        fig["layout"]["xaxis"]["title"] = "Cluster Labels"
        fig["layout"]["yaxis"]["title"] = "Cluster Mean Distances"
        del fig.layout["width"]
        del fig.layout["height"]
        return self.makeplot(fig, "agg/" + self.shortname)


class HierarchicalStackedClusterMeansHeatmap(MatrixPlotter):
    titlestring = "%s %s Stacked Cluster Means, Level %d"
    shortname = "scmh"

    def plot(self, showticklabels=False):
        title = self.titlestring % (self.DS.name, self.DS.clustname, self.DS.levels)
        self.shortname = self.DS.shortclustname + self.shortname

        Xs = []
        for l in self.DS.clusters[1:self.DS.levels + 1]:
            #When number of samples is too high, need to downsample
            freq = [c.shape[0] for c in l]
            if sum(freq) > 500:
                freq = [round((x / sum(freq)) * 500) for x in freq]
                if sum(freq) != 500: #Rounding can give numbers not exactly 500
                    freq[freq.index(max(freq))] += (500 - sum(freq))

            means = []
            for i, c in enumerate(l):
                means += [np.average(c, axis=0)] * freq[i]

            X = np.column_stack(means)
            Xs.append(X)
        X = np.vstack(Xs)[::-1, :]
        y_labels = np.tile(self.DS.columns, X.shape[0] // len(self.DS.columns))[::-1]
        trace = go.Heatmap(z = X,
                           zmin = -np.max(X),
                           zmax = np.max(X))
        data = [trace]
        xaxis = go.XAxis(
                title="Clusters",
                showticklabels=False,
                ticks="",
                mirror=True,
                tickvals = [i for i in range(X.shape[1])])
        yaxis = go.YAxis(
                title="Dimensions",
                showticklabels=showticklabels,
                ticks="",
                ticktext = y_labels,
                tickvals = [i for i in range(len(y_labels))],
                mirror=True)
        emb_size = len(np.average(self.DS.clusters[0][0], axis=0))
        bar_locations = np.arange(0, X.shape[0]  + emb_size - 1, emb_size) - 0.5
        shapes = [dict(type="line",x0=-0.5, x1=X.shape[1] - 0.5, y0=b, y1=b) for b in bar_locations]
        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis, shapes=shapes)
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)


class ClusterMeansLevelHeatmap(MatrixPlotter):
    titlestring = "%s %s Cluster Means, Level %d"
    shortname = "cmlh"

    def plot(self, showticklabels=False):
        title = self.titlestring % (self.DS.name, self.DS.clustname, self.DS.levels)
        self.shortname = self.DS.shortclustname + self.shortname

        #When number of samples is too high, need to downsample
        freq = [c.shape[0] for c in self.DS.clusters[self.DS.levels]]
        if sum(freq) > 500:
            freq = [round((x / sum(freq)) * 500) for x in freq]

        means = []
        for i, c in enumerate(self.DS.clusters[self.DS.levels]):
            means += [np.average(c, axis=0)] * freq[i]
        X = np.column_stack(means)

        trace = go.Heatmap(y = self.DS.columns,
                           z = X)
        data = [trace]
        xaxis = go.XAxis(
                title="Clusters",
                showticklabels=False,
                ticks="",
                mirror=True,
                tickvals = [i for i in range(X.shape[1])])
        yaxis = go.YAxis(
                title="Dimensions",
                showticklabels=showticklabels,
                ticks="",
                mirror=True)
        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)


class ClusterMeansLevelLines(MatrixPlotter):
    titlestring = "%s %s Cluster Means, Level %d"
    shortname = "cmll"

    def plot(self, showticklabels=False):
        title = self.titlestring % (self.DS.name, self.DS.clustname, self.DS.levels)
        self.shortname = self.DS.shortclustname + self.shortname
        data = []
        colors = get_spaced_colors(len(self.DS.clusters[self.DS.levels]))

        #When number of samples is too high, need to downsample
        freq = [c.shape[0] for c in self.DS.clusters[self.DS.levels]]
        if sum(freq) > 300:
            freq = [round((x / sum(freq)) * 300) for x in freq]

        for i, c in enumerate(self.DS.clusters[self.DS.levels]):
            data.append(go.Scatter(x = np.average(c, axis=0),
                                   y = self.DS.columns,
                                   mode="lines",
                                   line=dict(width=np.sqrt(freq[i]), color=colors[i]),
                                   name="cluster " + str(i)))
        xaxis = go.XAxis(
                title="Mean Values",
                showticklabels=False,
                mirror=True)
        yaxis = go.YAxis(
                title="Dimensions",
                showticklabels=showticklabels,
                mirror=True)
        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)


class ClusterPairsPlot(MatrixPlotter):
    titlestring = "%s %s Classification Pairs Plot, Level %d"
    shortname = "cpp"

    def plot(self):
        title = self.titlestring % (self.DS.name, self.DS.clustname, self.DS.levels)
        self.shortname = self.DS.shortclustname + self.shortname
        data = []
        colors = get_spaced_colors(len(self.DS.clusters[self.DS.levels]))
        samples = []
        labels = []
        for i, c in enumerate(self.DS.clusters[self.DS.levels]):
            samples.append(c.T)
            labels.append(c.shape[0] * [i])
        samples = np.hstack(samples)[:3, :]
        labels = np.hstack(labels)
        df = pd.DataFrame(samples.T, columns=["Dim %d"%i for i in range(samples.shape[0])])
        df["label"] = ["Cluster %d"%i for i in labels]
        fig = ff.create_scatterplotmatrix(df, diag='box', index="label", colormap=colors)
        fig["layout"]["title"] = title
        del fig.layout["width"]
        del fig.layout["height"]
        return self.makeplot(fig, "agg/" + self.shortname)


class DistanceMatrixPlotter:
    """A generic aggregate plotter acting on a distance matrix to be extended.

    Parameters
    ----------
    dm : :obj:`DistanceMatrix`
        The distance matrix object.
    primary_label : string
        The name of the column of the dataset which contains the primary label. By default, this is the `resource_path` column which is just the path to the data point resource.
    Attributes
    ----------
    dataset_name : string
        The name of the dataset from which this distance matrix was computed.
    dm : :obj:`ndarray`
        The distance matrix.
    label_name : string
        The name of the primary label to be conditioned on in some plots.
    label : :obj:`list`
        A list of labels (the primary label) for each data point.
    metric_name : string
        The name of the metric which with the distance matrix was computed.

    """

    def __init__(self, dm, mode = "notebook", primary_label = "resource_path"):
        self.dataset_name = dm.dataset.name
        self.dm = dm.getMatrix()
        self.labels = dm.labels
        self.label_name = dm.label_name
        self.metric_name = dm.metric.__name__
        self.plot_mode = mode

    def makeplot(self, fig):
        """Make the plotly figure visable to the user in the way they want.

        Parameters
        ----------
        gid : :obj:`figure`
            An plotly figure.

        """

        if self.plot_mode == "notebook":
            iplot(fig)
        if self.plot_mode == "html":
            fig["layout"]["autosize"] = True
            h = random.getrandbits(128)
            fname = "%032x.html"%h
            plot(fig, output_type='file', filename=fname)

class CSVPlotter:
    def __init__(self, ds, mode = "notebook"):
        self.ds = ds
        self.plot_mode = mode

    def makeplot(self, fig):
        """Make the plotly figure visable to the user in the way they want.

        Parameters
        ----------
        gid : :obj:`figure`
            An plotly figure.

        """

        if self.plot_mode == "notebook":
            iplot(fig)
        if self.plot_mode == "html":
            fig["layout"]["autosize"] = True
            h = random.getrandbits(128)
            fname = "%032x.html"%h
            plot(fig, output_type='file', filename=fname)
        if self.plot_mode == "div":
            fig["layout"]["autosize"] = True
            return plot(fig, output_type='div', include_plotlyjs=False)


class ColumnDistributionPlotter(CSVPlotter):
    def plot(self, column):
        x, y = self.ds.getColumnDistribution(column)
        yn = y / np.nansum(y)
        # title = "Column Distribution Plot<br>" + self.ds.getColumnDescription(column, sep="<br>")
        title = "Column Distribution Plot<br>"
        trace_frequency = go.Bar(
            x = x,
            y = y,
            name = 'Frequency'
        )
        trace_proportion = go.Bar(
            x = x,
            y = yn,
            visible = False,
            name = 'Proportion'
        )
        layout = go.Layout(
            title = title,
            xaxis = dict(title="Value"),
            yaxis = dict(title="Frequency")
        )
        updatemenus = list([
            dict(buttons = list([
                dict(args = [{'visible': [True, False]}, {'yaxis': dict(title="Frequency")}],
                     label = 'Frequency',
                     method = 'update'
                ),
                dict(args = [{'visible': [False, True]}, {'yaxis': dict(title="Proportion")}],
                     label = 'Proportion',
                     method = 'update'
                ),
                ]),
                showactive = True,
                type = 'buttons'
            )
        ])
        layout.updatemenus = updatemenus
        fig = go.Figure(data = [trace_frequency, trace_proportion], layout=layout)
        return self.makeplot(fig)

class ColumnNADistPlotter(CSVPlotter):
    def plot(self, column):
        na, not_na = self.ds.getColumnNADist(column)
        title = "Column NA Distribution Plot<br>" + self.ds.getColumnDescription(column, sep="<br>")
        trace = go.Pie(
            labels = ['NA', 'Not NA'],
            values = [na, not_na]
        )
        layout = go.Layout(
            title=title
        )
        fig = go.Figure(data = [trace], layout=layout)
        return self.makeplot(fig)

class EverythingPlotter(CSVPlotter):
    html_data = """
	<html>
	    <head>
		<title>
		%s
		</title>
		<script src="https://cdn.plot.ly/plotly-latest.min.js"></script>
	    </head>
	    <body>
              %s
            </body>
         </html>
	"""
    def plot(self, base, plotter):
        cp = plotter(self.ds, mode="div")
        for c in self.ds.D.columns:
            path = os.path.join(base, *c)
            os.makedirs(path, exist_ok=True)
            div = cp.plot(c)
            with open(os.path.join(path, plotter.__name__ + ".html"), "w") as f:
                f.write(self.html_data%(plotter.__name__, div))

class EmbeddingParallelCoordinatePlotter(DistanceMatrixPlotter):
    titlestring = "%s-%s-Embedding-Parallel-Coordinate-Plot-under-%s-metric"

    def plot(self, embedder):
        """Constructs a parallel coordinate plot of the embedded :obj:`DistanceMatrix` object.

        Parameters
        ----------
        embedder : :obj:`BaseEmbedder`


        """
        title = self.titlestring % (self.dataset_name, embedder.embedding_name, self.metric_name)
        emb = embedder.embed(self.dm)
        D = emb.T
        d, n = D.shape
        D = D - np.min(D, axis=1).reshape(d, 1)
        D = D / np.max(D, axis=1).reshape(d, 1)
        unique_labels = np.unique(self.labels)
        label_to_number = dict(zip(unique_labels, range(1, len(unique_labels) + 1)))
        dims = [dict(label = "factor %s"%(x + 1),
                values = D[x, :]) for x in range(embedder.num_components)]
        line = dict(color = [label_to_number[x] for x in self.labels],
                    cmin = 0,
                    cmax = len(unique_labels),
                    colorscale = "Jet",
                    showscale=True,
                    colorbar = dict(tickmode = "array",
                                    ticktext = unique_labels,
                                    tickvals = [label_to_number[x] for x in unique_labels]))
        trace = go.Parcoords(line = line, dimensions = list(dims))
        data = [trace]
        layout = go.Layout(
            title=title
        )
        fig = dict(data = data, layout = layout)
        self.makeplot(fig)




class SparkLinePlotter(TimeSeriesPlotter):
    titlestring = "Sparklines for %s"
    shortname = "sparkline"

    def plot(self, sample_freq):
        """Constructs a downsampled spark line plot of the time series.

        If there are more than 500 time points, the time series will be down sampled to
        500 column variables by windowed averaging. This is done by splitting the time series
        into 500 equal sized segments in the time domain, then plotting the mean for each segment.

        Parameters
        ----------
        sample_freq : int
            The sampling frequency (how many times sampled per second).

        """
        title = self.titlestring % (self.resource_name)
        xaxis = dict(
            title = "Time in Seconds"
        )
        yaxis = dict(
            title = "Intensity"
        )
        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)
        if self.n > 500:
            winsize = self.n // 500
            df = pd.DataFrame(self.data.T)
            df = df.groupby(lambda x: x // winsize).mean()
            downsampled_data = df.as_matrix().T
            data = [dict(mode="lines",
                         name = str(i),
                         x=(np.arange(downsampled_data.shape[1]) * winsize) / sample_freq,
                         y=downsampled_data[i, :]) for i in range(downsampled_data.shape[0])]
        fig = dict(data=data, layout=layout)
        self.makeplot(fig, self.resource_name + "/" + self.shortname)

class CorrelationMatrixPlotter(TimeSeriesPlotter):
    titlestring = "Correlation Matrix for %s"

    def plot(self):
        title = self.titlestring % (self.resource_name)
        xaxis = dict(
            title = "Channels"
        )
        yaxis = dict(
                scaleanchor="x",
            title = "Channels"
        )
        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)
        with np.errstate(divide = 'ignore', invalid = 'ignore'):
            C = np.nan_to_num(np.corrcoef(self.data))

        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)
        trace = go.Heatmap(z = C)
        fig = dict(data=[trace], layout=layout)
        self.makeplot(fig)

class CoherenceMatrixPlotter(TimeSeriesPlotter):
    titlestring = "Coherence Matrix for %s"

    def plot(self, samp_freq = 500):
        title = self.titlestring % (self.resource_name)
        xaxis = dict(
            title = "Channels"
        )
        yaxis = dict(
                scaleanchor="x",
            title = "Channels"
        )
        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)
        C = np.zeros([self.d, self.d])
        for i in range(self.d):
            for j in range(i + 1):
                C[i, j] = np.mean(np.nan_to_num(signal.coherence(self.data[i, :],
                                                              self.data[j, :],
                                                              fs=samp_freq)[1]))
                C[j, i] = C[i, j]

        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)
        trace = go.Heatmap(z = C)
        fig = dict(data=[trace], layout=layout)
        self.makeplot(fig)

class SpectrogramPlotter(TimeSeriesPlotter):
    titlestring = "Spectrograms for %s"

    def plot(self, channel = 0, sample_freq = 500):
        """Constructs a spectrogram plot of the time series.

        Parameters
        ----------
        sample_freq : int
            The sampling frequency (how many times sampled per second).

        """
        title = self.titlestring % (self.resource_name)
        xaxis = dict(
            title = "Hz",
            range = [0, 10]
        )
        yaxis = dict(
            title = "Intensity",
            range = [0, 1000]
        )
        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)

        dt = 1./sample_freq
        sample_points = np.arange(self.data.T.shape[1]) * dt
        signal = self.data.T[channel, :]
        ft = np.fft.fft(signal) * dt
        ft = ft[: len(sample_points)//2]
        freq = np.fft.fftfreq(len(sample_points), dt)
        freq = freq[:len(sample_points)//2]

        trace = go.Scatter(
            x = freq[:2000],
            y = np.abs(ft)[:2000],
            mode = 'markers'
        )

        fig= dict(data=[trace], layout=layout)
        iplot(fig)

class SpatialTimeSeries(TimeSeriesPlotter):
    titlestring = "Intensity by Time and Channel Location for %s"
    shortname = "spatialtimeseries"

    def plot(self, spatial, downsample = 1000):
        """Constructs a plot of the channel locations, and their intensities at different times.

        Parameters
        ----------
        spatial : Dataset
            Locations of channels.

        """
        title = self.titlestring % (self.resource_name)
        # Time series containing EEG data.
        mts = self.data.T

        # Set variables
        num_obs = mts.shape[0]
        num_channels = mts.shape[1]

        # Verify that 'locations' exist for exactly each channel.
        if (num_channels != spatial.shape[0]):
            raise TypeError("""Error: Ensure that the number of channels in the Multivariate Time Series (columns)
                            is equal to the number of points (rows) in locations.""")

        # Sets up data frame containing the different plots.
        data = [go.Scatter3d(dict(
            visible = False,
            name = str(step),
            x = spatial[:, 0],
            y = spatial[:, 1],
            z = spatial[:, 2],
            mode = 'markers',
            # Marker represents intensity.
            marker = dict(
            size = 8,
            color = mts[step, range(num_channels)],
            colorbar = go.ColorBar(
                    title='Voltage'
                ),
            colorscale='Viridis',
            line = dict(
                width = 1,
                color = 'rgb(0, 0, 0)'
            )))) for step in range(0, num_obs, downsample)]

        # Set up timesteps
        steps = []
        for i in range(len(data)):
            step = dict(
                method = 'restyle',
                args = ['visible', [False] * len(data)],
            )
            step['args'][1][i] = True # Toggle i'th trace to "visible"
            steps.append(step)

        # Sets up slider
        sliders = [dict(
            active = 10,
            currentvalue = {"prefix": "Timestep: "},
            pad = {"t": 50},
            steps = steps
        )]
        layout = dict(sliders=sliders)

        # Plot figure.
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, self.resource_name + "/" + self.shortname)

class SpatialPeriodogram(TimeSeriesPlotter):
    titlestring = "Density by Frequency and Channel Location for %s"
    shortname = "spatialpgram"

    def plot(self, spatial, downsample = 1000):
        """Constructs a plot of the channel locations, and their densities at different frequencies.

        Parameters
        ----------
        spatial : Dataset
            Locations of channels.

        """
        title = self.titlestring % (self.resource_name)
        # Time series containing EEG data.
        mts = self.data.T

        # Set variables
        num_obs = mts.shape[0]
        num_channels = mts.shape[1]

        # Create matriz to hold frequencies and densities
        freq0, density0 = signal.periodogram(mts[0:num_obs:downsample,0])
        densities = np.zeros((len(density0), num_channels))
        densities[:, 0] = density0
        for j in range(1, num_channels):
            freq, density = signal.periodogram(mts[0:num_obs:downsample,j])
            densities[:, j] = density
        num_dens = densities.shape[0]

        # Verify that 'locations' exist for exactly each channel.
        if (num_channels != spatial.shape[0]):
            raise TypeError("""Error: Ensure that the number of channels in the Multivariate Time Series (columns)
                            is equal to the number of points (rows) in locations.""")

        # Sets up data frame containing the different plots.
        data = [go.Scatter3d(dict(
            visible = False,
            name = 'Frequency = '+str(step),
            x = spatial[:, 0],
            y = spatial[:, 1],
            z = spatial[:, 2],
            mode = 'markers',
            # Marker represents density.
            marker = dict(
            size = 8,
            color = densities[step, range(num_channels)],
            colorbar = go.ColorBar(
                    title='Power Density'
                ),
            colorscale='Viridis',
            line = dict(
                width = 1,
                color = 'rgb(0, 0, 0)'
            )))) for step in range(num_dens)]

        # Set up timesteps
        steps = []
        for i in range(len(data)):
            step = dict(
                method = 'restyle',
                args = ['visible', [False] * len(data)],
            )
            step['args'][1][i] = True # Toggle i'th trace to "visible"
            steps.append(step)

        # Sets up slider
        sliders = [dict(
            active = 10,
            currentvalue = {"prefix": "Frequency: "},
            pad = {"t": 50},
            steps = steps
        )]
        layout = dict(sliders=sliders)

        # Plot figure.
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, self.resource_name + "/" + self.shortname)

class Nifti4DPlotter:
    name = "Nifti4DPlotter"
    def __init__(self, resource):
        self.path = resource[0]
        self.subject = resource[1]
        self.task = resource[2]

    def plot(self, downsample=1, out_base="."):
        out_path = os.path.join(out_base, self.subject, self.name, self.task)
        os.makedirs(out_path, exist_ok=True)
        raw = nib.load(self.path)
        M = np.max(raw.get_data())
        n = raw.shape[3]
        mean = nimage.mean_img(raw)
        xyzcuts = nilplot.find_xyz_cut_coords(mean)
        xcuts = nilplot.find_cut_slices(mean, "x")
        ycuts = nilplot.find_cut_slices(mean, "y")
        zcuts = nilplot.find_cut_slices(mean, "z")
        del raw
        nrange = range(0, n, downsample)
        for i, img in enumerate(nimage.iter_img(self.path)):
            if i in nrange:
                nilplot.plot_epi(nimage.math_img("img / %f"%(M), img=img),
                                  colorbar=False,
                                  output_file="%s/orth_epi%0d.png"%(out_path, i),
                                  annotate=True,
                                  cut_coords = xyzcuts,
                                  cmap="gist_heat")
                nilplot.plot_epi(nimage.math_img("img / %f"%(M), img=img),
                                  colorbar=False,
                                  output_file="%s/x_epi%0d.png"%(out_path, i),
                                  annotate=True,
                                  display_mode = "x",
                                  cut_coords = xcuts,
                                  cmap="gist_heat")
                nilplot.plot_epi(nimage.math_img("img / %f"%(M), img=img),
                                  colorbar=False,
                                  output_file="%s/y_epi%0d.png"%(out_path, i),
                                  annotate=True,
                                  display_mode = "y",
                                  cut_coords = ycuts,
                                  cmap="gist_heat")
                nilplot.plot_epi(nimage.math_img("img / %f"%(M), img=img),
                                  colorbar=False,
                                  output_file="%s/z_epi%0d.png"%(out_path, i),
                                  annotate=True,
                                  display_mode = "z",
                                  cut_coords = zcuts,
                                  cmap="gist_heat")


        slice_names = ["orth_epi", "x_epi", "y_epi", "z_epi"]
        for slic in slice_names:
            filenames = ["%s/%s%0d.png"%(out_path, slic, i) for i in nrange]
            with imageio.get_writer('%s/%s.gif'%(out_path, slic), mode='I') as writer:
                for i, filename in zip(nrange, filenames):
                    image = Image.open(filename)
                    draw = ImageDraw.Draw(image)
                    fnt = ImageFont.truetype('Pillow/Tests/fonts/FreeMono.ttf', 16)
                    draw.text((2, 2), str(i), font=fnt, fill=(255, 0, 0, 255))
                    image.save(filename, "PNG")
                    image = imageio.imread(filename)
                    writer.append_data(image)



import os
import boto3
import pandas as pd
import numpy as np
import pickle as pkl
import logging
import json
import glob

from nilearn import image as nimage
from nilearn import plotting as nilplot
import nibabel as nib

class DataSet:
    def __init__(self, D, name="default"):
        self.D = D
        self.n, self.d = self.D.shape
        self.name = name

    def getResource(self, index):
        return self.D.iloc[index, :]

    def saveMetaData(self, filepath):
        metadata = dict(d=self.d, n=self.n, name=self.name)
        string = json.dumps(metadata, indent=2)
        with open(filepath, 'w') as f:
            f.write(string)
        return string

    def getMatrix(self):
        return self.D.as_matrix()

class BIDSParser:

    def __init__(self, base_path):
        dataset_name = os.path.basename(os.path.normpath(base_path))
        dataset = {}
        subjects = [os.path.basename(x) for x in glob.glob(base_path + "/*")]
        if "chanlocs.csv" in subjects:
            subjects.remove("chanlocs.csv")
        if "metadata.json" in subjects:
            subjects.remove("metadata.json")
        print(base_path)
        print(subjects)
        for s in subjects:
            dataset.update({s:{}})
        for s in subjects:
            modalities = [os.path.basename(x) for x in glob.glob(os.path.join(base_path, s) + "/*")]
            print(modalities)
            for m in modalities:
                dataset[s].update({m:{}})
                files = [os.path.basename(x) for x in glob.glob(os.path.join(base_path, s, m) + "/*")]
                print(files)
                for f in files:
                    t = "".join(f.split("_")[1:]).split(".")[0]
                    dataset[s][m].update({t:f})
        self.dataset = dataset
        self.base_path = base_path
        print(self.dataset)

    def getModalityFrame(self, modality, extension):
        files = []
        subjects = []
        tasks = []
        for s in self.dataset.keys():
            for t in self.dataset[s][modality].keys():
                f = self.dataset[s][modality][t]
                if f.endswith(extension):
                    files.append(os.path.join(self.base_path, s, modality, f))
                    subjects.append(s)
                    tasks.append(t)
        d = {
            "resource_path": files,
            "subjects": subjects,
            "tasks": tasks        
        }
        return pd.DataFrame(d)

class fMRIDataSet:

    def __init__(self, dataframe_descriptor, name="fmri"):
        self.D = dataframe_descriptor
        self.D.index = self.D["subjects"] + "-" + self.D["tasks"]
        self.D.index.name = "index"
        self.name = name
        self.n = self.D.shape[0]

    def getResource(self, index):
        resource = self.D.ix[index]
        return resource

    def getMatrix(self, index):
        resource_path = self.D.ix[index][0]
        return nib.load(resource_path).get_data()

class EEGDataSet:

    def __init__(self, dataframe_descriptor, name="fmri"):
        self.D = dataframe_descriptor
        self.D.index = self.D["subjects"].astype(str) + "-" + self.D["tasks"].astype(str)
        self.D.index.name = "index"
        self.name = name
        self.n = self.D.shape[0]

    def getResource(self, index):
        resource = self.D.ix[index]
        return resource

    def getMatrix(self, index):
        resource_path = self.D.ix[index][0]
        with open(resource_path, "rb") as f:
            return pkl.load(f).T

    def getResourceDS(self, index):
        resource = self.getResource(index)
        matrix = self.getMatrix(index)
        D = pd.DataFrame(matrix.T)
        name = "%s/%s"%(resource[1], resource[2])
        DS = DataSet(D, name)
        return DS

class GraphDataSet:

    def __init__(self, dataframe_descriptor, name="fmri"):
        self.D = dataframe_descriptor
        self.D.index = self.D["subjects"].astype(str) + "-" + self.D["tasks"].astype(str)
        self.D.index.name = "index"
        self.name = name
        self.n = self.D.shape[0]

    def getResource(self, index):
        resource = self.D.ix[index]
        return resource

    def getMatrix(self, index):
        resource_path = self.D.ix[index][0]
        try:
            return nx.read_graphml(files)
        except:
            return nx.read_gpickle(files)


class DiskDataSet:
    """A dataset living locally on the hard disk.

    A disk data set is defined by a `.csv` file where entries of the `resource_path` column 
    are paths to local `.pkl` files, and all other columns describe variables of the data point 
    linked to by the `resource_path` variable. 

    Parameters
    ----------
    df_path : str
	Path to the .csv file describing the DiskDataSet.

    Attributes
    ----------
    D : pandas DataFrame
        A DataFrame object describing the dataset.
    N : int
        The number of observations in the dataset.
    name : string
        A descriptive name for the dataset.

    """

    def __init__(self, df_path, index_column = None):
        self.D = pd.read_csv(df_path)
        if index_column is not None:
            self.D.index = self.D[index_column]
            self.D.index.name = index_column
        self.n = self.D.shape[0]
        self.name = df_path.split("/")[-1].split(".")[0].split("_")[0]

    def getResource(self, index):
        """Get a specific data point from the data set.

        Parameters
        ----------
        index : int
            The index of the data point in `D`.

        Returns
        -------
        :obj:`ndarray`
            A ndarray of the data point.

        """
        resource_path = self.D["resource_path"].ix[index]
        dim_column = self.D["dim_column"].ix[index]
        with open(resource_path, "rb") as f:
            if dim_column:
                return pkl.load(f).T
            return pkl.load(f)

    def getResourceDS(self, index):
        D = pd.DataFrame(self.getResource(index).T)
        name = self.name + " " + \
               str(index) + " "
        DS = DataSet(D, name)
        return DS

class CloudDataSet:
    """A dataset living in the cloud in S3.

    A cloud data set lives in S3 with a connection defined by a `.csv` file which
    contains the parameters for a user's bucket.


    Parameters
    ----------
    df_path : str
	Path to the .csv file describing the CloudDataSet.

    Attributes
    ----------
    D : pandas DataFrame
        A DataFrame object describing the dataset.
    N : int
        The number of observations in the dataset.
    name : string
        A descriptive name for the dataset.

    """

    def __init__(self, df_path):
        credential_info = open(df_path, 'r').readlines()
        self.client = boto3.client(
            's3',
            aws_access_key_id=credential_info[1][:-1],
            aws_secret_access_key=credential_info[2],
        )
        self.bucket_name = credential_info[0][:-1]
        self.objects = self.client.list_objects(self.bucket_name)['Contents']

    def getResource(self, index):
        """Get a specific data point from the data set.

        Parameters
        ----------
        index : int
            The index of the data point in `D`.

        Returns
        -------
        :obj:`ndarray`
            A ndarray of the data point.

        """
        self.client.download_file(self.bucket_name, index)

    def getResourceDS(self, index):
        D = pd.DataFrame(self.getResource(index).T)
        name = self.name + " " + \
               str(index) + " "
        DS = DataSet(D, name)
        return DS

def convertDtype(l):
    try:
        return np.array(l, dtype="float")
    except:
        pass
    l = np.array(l, dtype=str)
    l[l == 'nan'] = 'NA'
    return l

class CSVDataSet(DataSet):
    """ A dataset living locally in a .csv file

    """
    def __init__(self, csv_path, index_column = None,  NA_val = ".", name = "mydataset"):
        self.name = name
        
        # Load the data set
        D = pd.read_csv(csv_path, dtype="unicode")
        self.n, self.d = D.shape
        print("Dataset of size", self.n, "samples", self.d, "dimensions", "Loaded")

        # Convert to numeric all numeric rows
        D = D.replace(NA_val, "nan")
        print("Replacing all", NA_val, "with nan")
        d = []
        for c in D.columns:
            d.append(convertDtype(list(D[c])))
            print("Converting", c, end="\r\r") 
        newcolumns = D.columns
        newindex = D.index
        D = list(d)
        D = pd.DataFrame(dict(zip(newcolumns, D)), index = newindex)


        # Set the index column as specified
        if index_column is not None:
            print("Setting index column as", index_column)
            D.index = D[index_column]
            print("Deleting", index_column, "from dataset")
            del D[index_column]

        self.D = D

        # Remove all columns which have all null values
        keep = []
        allnull = self.D.isnull().all(axis=0)
        for c in self.D.columns[allnull]:
            print("Removing column", c, "because it has all null values")
        keep = self.D.columns[~allnull]
        self.D = self.D[keep]

        # Remove all rows which have all null values
        allnull = self.D.isnull().all(axis=1)
        for r in self.D.index[allnull]:
            print("Removing row", r, "because it has all null values")
        keep = self.D.index[~allnull]
        self.D = self.D.loc[keep]
        n, d = self.D.shape
        print("Dataset of size", n, "samples", d, "dimensions", "Resulting")
        self.N = self.D.shape[0]

    def imputeColumns(self, numeric):
        keep = []
        keep = (self.D.dtypes == "float64").as_matrix()
        for c in self.D.columns[~keep]:
            print("Removing column", c, "because it is not numeric")
        self.D = self.D[self.D.columns[keep]]
        cmean = self.D.mean(axis=0)
        values = dict(list(zip(self.D.columns, cmean.as_matrix())))
        #self.D.fillna(value=values, inplace=True)
        d = self.D.as_matrix()
        for i, c in enumerate(self.D.columns):
            print("Imputing column", c, "with value", values[c])
            d[:, i][np.isnan(d[:, i])] = values[c]
        D = pd.DataFrame(d)
        D.index = self.D.index
        D.index.names = self.D.index.names
        D.columns = self.D.columns
        D.columns.names = self.D.columns.names
        self.D = D
        allzero = np.all(self.D.as_matrix() == 0, axis=0)
        for c in self.D.columns[allzero]:
            print("Removing column", c, "because it has all zero values")
        keep = self.D.columns[~allzero]
        allsame = np.std(self.D.as_matrix(), axis=0) == 0
        for c in self.D.columns[allsame]:
            print("Removing column", c, "because it has all zero standard deviation (all values same)")
        keep = self.D.columns[~allsame]
        self.D = self.D[keep]
        n, d = self.D.shape
        print("Dataset of size", n, "samples", d, "dimensions", "Resulting")
        print("Dataset has", self.D.isnull().sum().sum(), "nans")
        print("Dataset has", np.sum(np.isinf(self.D.as_matrix())), "infs")

    def getResource(self, index):
        """Get a specific data point from the data set.

        Parameters
        ----------
        index : int or string
            The index of the data point in `D`, either positional or a string.

        Returns
        -------
        :obj:`ndarray`
            A ndarray of the data point.

        """
        if type(index) is int:
            return self.D.iloc[index].as_matrix()
        else:
            return self.D.loc[index].as_matrix()

    def getColumn(self, index):
        """Get a column of the dataframe.
 
        Parameters
        ----------
        index : int or string
            The index of the column in `D`, either positional or a string.

        Returns
        -------
        :obj:`ndarray`
            The values in the column.
        """
        if type(index) is int:
            return self.D.iloc[:, index].as_matrix()
        else:
            return self.D[index].as_matrix()

    def getColumnValues(self, index):
        """Get the unique values of a column.

        Parameters
        ----------
        index : int or string
            The index of the column in `D`, either positional or a string.

        Returns
        -------
        :obj:`ndarray`
            A ndarray of the unique values.

        """
        column = self.getColumn(index)
        if column.dtype == "float64":
            column = column[~np.isnan(column)]
        else:
            column = column[np.array([x != "NA" for x in column])]
        return np.unique(column)


    def getColumnDistribution(self, index):
        """Get the distribution of values in a column.

        Parameters
        ----------
        index : int or string
            The index of the column in `D`, either positional or a string.

        Returns
        -------
        :obj:`ndarray`, :obj:`ndarray`
            An array x of the unique labels, and an array y of the count of that label

        """
        x = self.getColumnValues(index)
        column = self.getColumn(index)
        y = [np.sum(column == v) for v in x]
        return x, y

    def getColumnNADist(self, index):
        column = self.getColumn(index)
        if column.dtype == "float64":
            na = np.sum([np.isnan(x) for x in column])
            not_na = len(column) - na
            return na, not_na
        else:
            na = np.sum([x == "NA" for x in column])
            not_na = len(column) - na
            return na, not_na
        return na, not_na

    def getColumnDescription(self, index, sep = "\n"):
        """Get a description of the column.

        """
        desc = []
        if type(index) is int:
            index = self.D.columns.values[index]
        for i, name in enumerate(self.D.columns.names):
            desc.append(name + ": " + index[i])
        return sep.join(desc)

    def getLevelValues(self, index):
        return np.unique(self.D.columns.get_level_values(index))


class DistanceMatrix:
    """A distance matrix computed from a DataSet object.

    Parameters
    ----------
    dataset : :obj:`DiskDataSet`
        A dataset on which to compute the distance matrix
    metric : function
        A distance used to compute the distance matrix.

    Attributes
    ----------
    dataset : :obj:`DiskDataSet`
        A dataset on which to compute the distance matrix
    metric : function
        A distance used to compute the distance matrix.
    N : int
        Number of data points in the dataset.
    matrix : :obj:`ndarray`
        The distance matrix.

    """

    def __init__(self, dataset, metric):
        self.DS = dataset
        self.name = self.DS.name
        self.labels = self.DS.D.index.values
        self.label_name = self.DS.D.index.name
        self.metric = metric
        self.metric_name = metric.__name__
        self.n = self.DS.n
        parameterization = self.metric.parameterize(self.DS)
        self.D = np.zeros([self.n, self.n])
        for i in range(self.n):
            I = parameterization[i]
            for j in range(i + 1):
                J = parameterization[j]
                self.D[i, j] = self.metric.compare(I, J)
                self.D[j, i] = self.D[i, j]
        self.D = pd.DataFrame(self.D)
        self.D.index = self.DS.D.index
        self.D.index.name = self.DS.D.index.name

    def getMatrix(self):
        """Get the distance matrix.

        Returns
        -------
        :obj:`ndarray`
            The distance matrix.

        """
        return self.D

class InterpointMatrix:

    def __init__(self, dataset, metric):
        self.DS = dataset
        self.name = self.DS.name
        self.metric = metric
        self.metric_name = metric.__name__
        self.d = self.DS.d
        self.D = np.zeros([self.d, self.d])
        for i in range(self.d):
            I = self.dataset.D.iloc[:, i]
            for j in range(i + 1):
                J = self.DS.D.iloc[:, j]
                self.D[i, j] = self.metric.compare(I, J)
                self.D[j, i] = self.D[i, j]
        self.D = pd.DataFrame(self.D)
        self.D.index = self.DS.D.columns
        self.D.index.name = self.DS.D.index.name

    def getMatrix(self):
        """Get the distance matrix.

        Returns
        -------
        :obj:`ndarray`
            The distance matrix.

        """
        return self.D



from distutils.core import setup
setup(
  name = 'redlemur',
  packages = ['lemur', 'lemur.utils'],
  version = '0.28',
  description = 'High level visualization library',
  author = 'Nitin Kumar',
  author_email = 'nkumar14@jhu.edu',
  url = 'https://github.com/NeuroDataDesign/lemur',
  download_url = 'https://github.com/NeuroDataDesign/lemur/archive/0.28.tar.gz',
  keywords = ['visualization', 'neuroscience', 'embeddings'],
  classifiers = [],
)
